#!/usr/bin/env python

# Alsamixertest - tests that your alsa mixer lives up to PulseAudio's expectations
#
# Written by David Henningsson <david.henningsson@canonical.com>
#
#    Copyright 2010 Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify it 
#    under the terms of the GNU General Public License version 3, as published 
#    by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful, but 
#    WITHOUT ANY WARRANTY; without even the implied warranties of 
#    MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
#    PURPOSE.  See the GNU General Public License for more details.#
#
#    You should have received a copy of the GNU General Public License along 
#    with this program.  If not, see <http://www.gnu.org/licenses/>.

from mixercontrol import MixerControl, MixerControlList
import utils
import pulse_control
from utils import run_subprocess
import logging
from optparse import OptionParser
import shutil
import sys
import os.path

class Struct():
    pass

def runtest(in_device, out_device):
    ''' Runs a test signal iteration, returns array of (freq, dB) tuples '''
    logging.debug("Running test...")
    run_subprocess(("./testcase.sh", in_device, out_device))
    f = open("/tmp/dbtest.txt", "r")
    result = []
    for line in f:
        s = line.split(" ")
        result.append((int(s[0]), float(s[2])))
    return result

def analyzetestresult(freqdata):
    ''' Input: (freq, dB) tuples from runtest '''
    ''' Output: Struct '''
    if (len(freqdata) < 10):
        logging.critical("Invalid test result data")
        return Struct()
    result = Struct()
    expectedrate = 440
    # print freqdata
    actualrate, actuallevel = freqdata[0]
    noiserate, noiselevel = freqdata[1]

    if (abs(actualrate - expectedrate) > 3):
        if (actuallevel - 30 > noiselevel):
            logging.warning("Signal sample rate mismatch: {0}, expected {1}! " \
                "".format(actualrate, signalrate))
            # should we continue anyway and accept this rate deviation?
        result.signal = False
        result.noise = actuallevel
        logging.debug("No signal, noise: {0} dB".format(result.noise))
    else:
        result.signal = True
        result.level = actuallevel
        result.noise = actuallevel - noiselevel
        if (abs(noiserate - actualrate*3) < 3):
            result.distortion = result.noise
            logging.debug("Distortion level: {0} dB".format(result.distortion))
        else:
            result.distortion = False
        logging.debug("Signal at {0} Hz, Level: {1} dB, S/N: {2} dB, " \
            "Noise: {3} dB". \
            format(actualrate, actuallevel, result.noise, noiselevel))
    return result

  
def checkreference(level):
    if not level.signal:
        logging.error("No signal detected. Check that the loopback is " \
            "correctly set up and that capture levels are correctly set. ")
        logging.error("Error 1: Playback mixers present in signal "\
            "path with non-standard mixer names.")
        return False
    if level.distortion and level.distortion < 60:
        logging.error("Distortion/clipping detected. Try reducing capture "\
            "levels and try again.")
        logging.error("Error 2: Setting all mixers in signal "\
            "path causes digital clipping.")
        return False
    if level.noise < 10:
        logging.warn("Bad signal quality, noise is at {0:.4f}dB". \
            format(level.noise))
    return True


def runtestsuite(devices, expected_path, dbstep):

    expected_path = set(expected_path)
    controls = MixerControlList(devices["mixer"])
    volumecontrols = controls.get_control_cap(['pvolume', 'pvolume-joined', 'volume']);
    switchcontrols = controls.get_control_cap(['pswitch', 'switch']);

    # set up mixer for initial test run
    if expected_path.isdisjoint(c.get_pretty_name() for c in volumecontrols):
        logging.error("Error 9: No volume controls in standard path.")
    for c in volumecontrols:
        if c.has_dB:
            if c.get_pretty_name() in expected_path:
                c.set_dB(0)
            else:
                c.set_dB(-99999)
        else:
            logging.error("Error 4: Volume {0} does not "\
                "support getting/setting dB information.".
                format(c.get_pretty_name()))

    for c in switchcontrols:
        c.set_mute(not c.get_pretty_name() in expected_path) 

    # run initial test
    logging.info("Running initial test signal.")
    q = runtest(devices["input"],devices["output"])
    referencelevel = analyzetestresult(q)
    if not checkreference(referencelevel):
        return False

    # run mute test
    for c in switchcontrols:
        if not c.get_pretty_name() in expected_path:
            continue
        logging.info("Testing that {0} actually mutes the signal".format(
            c.get_pretty_name()))
        c.set_mute(True) 
        m = analyzetestresult(runtest(devices["input"],devices["output"]))
        if m.signal:
            logging.error("Error 3: Mute switch {0} does not "\
                "mute the signal, still found signal at {1:.3f}dB".format( 
                c.get_pretty_name(), m.level))
        c.set_mute(False)

    # run dB test
    logging.debug("Testing that dB data is correct for playback volume")
    for c in volumecontrols:
        if not c.get_pretty_name() in expected_path:
            continue
        if not c.has_dB:
            continue

        expected_noise = referencelevel.level-referencelevel.noise
        prev_real_dB = 999999
        cur_dB = 0.0
        while True:
            cur_dB -= dbstep # 2 dB steps
            c.set_dB(cur_dB)
            pstate = c.get_pstate()
            real_dB = sum(x['dB'] for x in pstate.itervalues()) / len(pstate)
            logging.debug("dB = {1} gives pstate = {0}".format(pstate, cur_dB))
            if real_dB == prev_real_dB:
                continue
            prev_real_dB = real_dB
            expected_level = referencelevel.level + real_dB
            if (expected_level <= expected_noise):
                logging.debug("Expected noise {1:.2f}dB, expected level "\
                    "{1:.2f}dB - breaking.".format(expected_noise, expected_level))
                logging.warning("Not testing mixer {0} at {1:.2f}dB and below, due to "\
                    "background noise.".format(c.get_pretty_name(), real_dB))
                break

            logging.info("Running test for mixer {0}, level {1:.2f}dB".format(
                c.get_pretty_name(), real_dB))
            m = analyzetestresult(runtest(devices["input"],devices["output"]))
            if not m.signal:
                logging.error("Error 5: Mixer {0} mutes "\
                    "the signal at {1:.2f}dB.".format(c.get_pretty_name(), cur_dB))
                break
            logging.debug("Measured level: {0}, expected {1}".format(m.level, expected_level))
            diff = expected_level - m.level
            if abs(diff) > 1:
                if (abs(diff) > 5):
                    logging.error("Error 6: Mixer {0} has invalid dB "\
                        "data for dB={1}, expected {2:.2f}dB but measured {3:.2f}dB.".format(
                        c.get_pretty_name(), real_dB, expected_level, m.level))
                else:
                    logging.warning("Mixer {0} is {1:.2f}dB off at {2:.2f} dB."\
                        .format(c.get_pretty_name(), diff, real_dB))
            else:
                expected_noise = m.level-m.noise

            if sum(x['value'] for x in pstate.itervalues()) <= 0:
                logging.debug("Measured all levels.")
                break
        c.set_dB(0)
    
def handle_argv():
    ''' Returns a tuple of (devices, selected_path) '''

    # Taken from /usr/share/pulseaudio/alsa-mixer/paths
    standard_paths = {
        'analog-output': set(['PCM', 'Hardware Master', 'Master', 
            'Master Front', 'Front']),
        'analog-output-headphones': set(['PCM', 'Hardware Master', 'Master', 
            'Master Front', 'Headphone']),
        'analog-output-speaker-front': set(['PCM', 'Hardware Master', 'Master', 
            'Master Front', 'Front', 'Speaker']),
        'analog-output-speaker-rear': set(['PCM', 'Hardware Master', 'Master', 
            'Rear', 'Surround', 'Speaker']),
        'analog-output-speaker-side': set(['PCM', 'Hardware Master', 'Master', 
            'Side', 'Speaker']),
        'analog-output-speaker-center': set(['PCM', 'Hardware Master', 'Master', 
            'Center', 'Speaker']),
        'analog-output-speaker-lfe': set(['PCM', 'Hardware Master', 'Master', 
            'LFE', 'Speaker'])
    }
    
    # Taken from /usr/share/pulseaudio/alsa-mixer/profile-sets/default.conf    
    device_prefix = {
        'analog-output': 'plug:front:',
        'analog-output-headphones': 'plug:front:',
        'analog-output-speaker-front': 'plug:front:',
        'analog-output-speaker-rear': 'plug:surround51:',
        'analog-output-speaker-side': 'plug:surround71:',
        'analog-output-speaker-center': 'plug:surround51:',
        'analog-output-speaker-lfe': 'plug:surround51:'
    }
    
    parser = OptionParser()
    parser.add_option("-c", "--sound-card", metavar="card", 
        dest="c", default="0",
        help="Sets up a loopback through the specified sound card."\
        " Available options: {0}".format(", ".join(utils.get_asound_names())))
    parser.add_option("-r", "--readme", action="store_true", dest="readme",
        help="View the readme file", default=False)
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose",
        help="Verbose debug information", default=False)
    parser.add_option("-i", "--input-device", metavar="device", 
        dest="input",
        help="ALSA device name for input device (overrides the -c switch)")
    parser.add_option("-o", "--output-device", metavar="device", 
        dest="output",
        help="ALSA device name for output device (overrides the -c switch)")
    parser.add_option("-m", "--mixer-device", metavar="device", 
        dest="mixer",
        help="ALSA device name for mixer device (overrides the -c switch)")
    parser.add_option("-s", "--step", metavar="value", dest="step",
        default=2,
        help="Step size (dB) - lower value means more testing levels")

    l = list(standard_paths.iterkeys())
    l.sort()
    parser.add_option("-p", "--path", metavar="path", 
        dest="path", default="analog-output",
        help="PulseAudio mixer path to test, available options:\n" +
            ", ".join(l)
            )

    (options, args) = parser.parse_args()

    # make sure we can access README.txt, the sine wave, and all the others
    (p, dummy) = os.path.split(os.path.realpath(__file__))
    os.chdir(p)

    if options.readme:
        f = open("README.txt", "r")
	print f.read()
        sys.exit(0)

    devices = {"input": "plughw:" + options.c, 
        "output": device_prefix[options.path] + options.c,  
        "mixer": "hw:" + options.c }
    if options.input: devices['input'] = options.input
    if options.output: devices['output'] = options.output
    if options.mixer: devices['mixer'] = options.mixer
    if options.verbose: 
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    step=float(options.step)

    selected_path = standard_paths[options.path]
    return (devices, selected_path, step)

def main():
    (devices, selected_path, dbstep) = handle_argv()

    s = pulse_control.getpulseaudio_streams()
    if (s is not None) and (len(s) > 0):
        logging.error("These applications are currently playing/recording audio and will " +
             "interfere with the test: {0} Close them and try again.".format(s))
        return

    pulse_control.killpulseaudio()
    try:
        savedcontrols = MixerControlList(devices["mixer"])
        try:
            runtestsuite(devices, selected_path, dbstep)
        finally:
            savedcontrols.restore_all() 
    finally:
        pulse_control.restorepulseaudio()



''' Main '''
if __name__ == '__main__':
    main()




